@@ -1,12 +1,12 @@
 import platform
 import shutil
 import stat
+import time
 from contextlib import suppress
 from datetime import datetime
 from os import DirEntry, getcwd, lstat, makedirs, path, remove, walk
 from typing import ClassVar
 
-from humanize import naturalsize
 from rich.segment import Segment
 from rich.style import Style
 from send2trash import send2trash
@@ -22,14 +22,16 @@ from textual.types import UnusedParameter
 from textual.widgets import Label, ProgressBar, SelectionList, Static
 from textual.widgets.option_list import OptionDoesNotExist
 from textual.widgets.selection_list import Selection
+from textual.worker import WorkerState
 
 from . import utils
+from .maps import SPINNER
 from .ScreensCore import CopyOverwrite, Dismissable, YesOrNo
 from .utils import config
 
 
 class ClipboardSelection(Selection):
-    def __init__(self, prompt: ContentText, *args, **kwargs):
+    def __init__(self, prompt: ContentText, *args, **kwargs) -> None:
         """
         Initialise the selection.
 
@@ -78,7 +80,7 @@ class Clipboard(SelectionList, inherit_bindings=False):
         ]
     )
 
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs) -> None:
         super().__init__(**kwargs)
         self.clipboard_contents = []
 
@@ -227,7 +229,7 @@ class Clipboard(SelectionList, inherit_bindings=False):
             if event.key in config["keybinds"]["delete"]:
                 """Delete the selected files from the clipboard."""
                 if self.highlighted is None:
-                    self.app.notify(
+                    self.notify(
                         "No files selected to delete from the clipboard.",
                         title="Clipboard",
                         severity="warning",
@@ -243,18 +245,21 @@ class Clipboard(SelectionList, inherit_bindings=False):
 
 
 class MetadataContainer(VerticalScroll):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
         self.current_path: str | None = None
         self._size_worker = None
         self._update_task = None
+        self._queued_task = None
+        self._queued_task_args: None | DirEntry = None
 
     def info_of_dir_entry(self, dir_entry: DirEntry, type_string: str) -> str:
-        """
-        Get the permission line from a given DirEntry object
+        """Get the permission line from a given DirEntry object
         Args:
             dir_entry (DirEntry): The nt.DirEntry class
             type_string (str): The type of file. It should already be handled.
+        Returns:
+            str: A permission string.
         """
         try:
             file_stat = lstat(dir_entry.path)
@@ -273,7 +278,7 @@ class MetadataContainer(VerticalScroll):
             case "File":
                 permission_string = "-"
             case "Unknown":
-                return "????????"
+                return "???????"
 
         permission_string += "r" if mode & stat.S_IRUSR else "-"
         permission_string += "w" if mode & stat.S_IWUSR else "-"
@@ -288,32 +293,44 @@ class MetadataContainer(VerticalScroll):
         permission_string += "x" if mode & stat.S_IXOTH else "-"
         return permission_string
 
-    @work(exclusive=True)
-    async def update_metadata(self, dir_entry: DirEntry) -> None:
+    def any_in_queue(self) -> bool:
+        if self._queued_task is not None:
+            self._queued_task(self._queued_task_args)
+            self._queued_task, self._queued_task_args = None, None
+            return True
+        return False
+
+    def update_metadata(self, dir_entry: DirEntry) -> None:
         """
         Debounce the update, because some people can be speed travellers
         Args:
             dir_entry (DirEntry): The nt.DirEntry object
         """
-        if self._update_task:
-            self._update_task.stop()
-        self._update_task = self.set_timer(
-            0.25, lambda: self._perform_update(dir_entry)
-        )
+        if any(
+            worker.is_running
+            and worker.node is self
+            and worker.name == "_perform_update"
+            for worker in self.app.workers
+        ):
+            self._queued_task = self._perform_update
+            self._queued_task_args = dir_entry
+        else:
+            self._perform_update(dir_entry)
 
-    async def _perform_update(self, dir_entry: DirEntry) -> None:
+    @work(thread=True)
+    def _perform_update(self, dir_entry: DirEntry) -> None:
         """
         After debouncing the update
         Args:
             dir_entry (DirEntry): The nt.DirEntry object
         """
-        if self._size_worker:
-            self._size_worker.cancel()
-            self._size_worker = None
-
+        if self.any_in_queue():
+            return
         if not path.exists(dir_entry.path):
-            await self.remove_children()
-            await self.mount(Static("Item not found or inaccessible."))
+            self.app.call_from_thread(self.remove_children)
+            self.app.call_from_thread(
+                self.mount, Static("Item not found or inaccessible.")
+            )
             return
 
         type_str = "Unknown"
@@ -338,7 +355,7 @@ class MetadataContainer(VerticalScroll):
                 case "size":
                     values_list.append(
                         Static(
-                            naturalsize(file_stat.st_size)
+                            utils.natural_size(file_stat.st_size)
                             if type_str == "File"
                             else "--",
                             id="metadata-size",
@@ -368,70 +385,100 @@ class MetadataContainer(VerticalScroll):
                             )
                         )
                     )
-
+        if self.any_in_queue():
+            return
         values = VerticalGroup(*values_list, id="metadata-values")
 
         try:
-            await self.query_one("#metadata-values").remove()
-            await self.mount(values)
+            for index, child_widget in enumerate(
+                self.query_one("#metadata-values").children
+            ):
+                self.app.call_from_thread(
+                    child_widget.update, values_list[index]._content
+                )
         except NoMatches:
-            await self.remove_children()
+            self.app.call_from_thread(self.remove_children)
             keys_list = []
             for field in config["metadata"]["fields"]:
-                if field == "type":
-                    keys_list.append(Static("Type"))
-                elif field == "permissions":
-                    keys_list.append(Static("Permissions"))
-                elif field == "size":
-                    keys_list.append(Static("Size"))
-                elif field == "modified":
-                    keys_list.append(Static("Modified"))
-                elif field == "accessed":
-                    keys_list.append(Static("Accessed"))
-                elif field == "created":
-                    keys_list.append(Static("Created"))
+                match field:
+                    case "type":
+                        keys_list.append(Static("Type"))
+                    case "permissions":
+                        keys_list.append(Static("Permissions"))
+                    case "size":
+                        keys_list.append(Static("Size"))
+                    case "modified":
+                        keys_list.append(Static("Modified"))
+                    case "accessed":
+                        keys_list.append(Static("Accessed"))
+                    case "created":
+                        keys_list.append(Static("Created"))
             keys = VerticalGroup(*keys_list, id="metadata-keys")
-            await self.mount(keys, values)
+            self.app.call_from_thread(self.mount, keys, values)
+        finally:
+            if self.any_in_queue():
+                return
         self.current_path = dir_entry.path
         if type_str == "Directory" and self.has_focus:
             self._size_worker = self.calculate_folder_size(dir_entry.path)
+        if self.any_in_queue():
+            return
+        else:
+            self._queued_task = None
 
-    @work
+    @work(thread=True)
     async def calculate_folder_size(self, folder_path: str) -> None:
         """Calculate the size of a folder and update the metadata."""
         size_widget = self.query_one("#metadata-size", Static)
-        self.call_later(size_widget.update, "Calculating...")
+        self.app.call_from_thread(size_widget.update, "Calculating...")
 
         total_size = 0
+        spinner_index = -1
+        last_update_time = time.monotonic()
         try:
             for dirpath, _, filenames in walk(folder_path):
-                if self._size_worker.is_cancelled:
-                    self.call_later(size_widget.update, "--")
-                    return
                 for f in filenames:
+                    if self._size_worker is None or self._size_worker.is_cancelled:
+                        return
                     fp = path.join(dirpath, f)
                     if not path.islink(fp):
                         with suppress(OSError, FileNotFoundError):
                             total_size += lstat(fp).st_size
+                if time.monotonic() - last_update_time > 0.25:
+                    spinner_index = spinner_index + 1 if spinner_index // 3 == 0 else 0
+                    self.app.call_from_thread(
+                        size_widget.update,
+                        f"{utils.natural_size(total_size)} {SPINNER[spinner_index]}",
+                    )
+                    last_update_time = time.monotonic()
         except (OSError, FileNotFoundError):
-            self.call_later(size_widget.update, "Error")
+            self.app.call_from_thread(size_widget.update, "Error")
             return
 
         if not self._size_worker.is_cancelled:
-            self.call_later(size_widget.update, naturalsize(total_size))
+            self.app.call_from_thread(
+                size_widget.update, utils.natural_size(total_size)
+            )
 
     @on(events.Focus)
     def on_focus(self) -> None:
         if self.current_path and path.isdir(self.current_path):
             if self._size_worker:
-                self._size_worker.cancel()
+                return
             self._size_worker = self.calculate_folder_size(self.current_path)
 
     @on(events.Blur)
     def on_blur(self) -> None:
-        if self._size_worker:
+        if self._size_worker is None or self.app.app_blurred:
+            return
+        elif self._size_worker.state == WorkerState.SUCCESS:
+            self._size_worker = None
+        else:
             self._size_worker.cancel()
             self._size_worker = None
+            self.set_timer(
+                0.1, lambda: self.query_one("#metadata-size", Static).update("--")
+            )
 
 
 class ProgressBarContainer(VerticalGroup):
@@ -442,7 +489,7 @@ class ProgressBarContainer(VerticalGroup):
         gradient: Gradient | None = None,
         *args,
         **kwargs,
-    ):
+    ) -> None:
         super().__init__(*args, **kwargs)
         self.progress_bar = ProgressBar(
             total=total,
@@ -476,7 +523,7 @@ class ProgressBarContainer(VerticalGroup):
 
 
 class ProcessContainer(VerticalScroll):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs) -> None:
         super().__init__(id="processes", *args, **kwargs)
 
     async def new_process_bar(
@@ -535,7 +582,26 @@ class ProcessContainer(VerticalScroll):
                                 path_to_trash = path_to_trash[6:]
                             send2trash(path_to_trash)
                         except FileNotFoundError:
-                            self.app.notify("FileNotFoundError")
+                            if path.exists(path_to_trash):
+                                # edge case: have no write access
+                                msg = (
+                                    "Trashing failed due to no write access.\nContinue?"
+                                )
+                            else:
+                                # really isn't found
+                                msg = f"{path_to_trash} could not be found.\nContinue?"
+                            do_continue = self.app.call_from_thread(
+                                self.app.push_screen_wait, YesOrNo(msg)
+                            )
+                            if do_continue:
+                                continue
+                            else:
+                                self.app.call_from_thread(
+                                    bar.update_label,
+                                    f"{utils.get_icon('general', 'delete')[0]} {utils.get_icon('general', 'close')[0]} Process Cancelled",
+                                )
+                                self.app.call_from_thread(bar.add_class, "error")
+                                return
                         except Exception as e:
                             perma_delete = self.app.call_from_thread(
                                 self.app.push_screen_wait,
@@ -585,7 +651,7 @@ class ProcessContainer(VerticalScroll):
                 shutil.rmtree(folder)
             except PermissionError:
                 # TODO: allow continuation and not return on error
-                self.app.notify(
+                self.notify(
                     f"Certain files in {folder} could not be deleted.", severity="error"
                 )
                 self.app.call_from_thread(
@@ -599,7 +665,7 @@ class ProcessContainer(VerticalScroll):
         if files_to_delete == []:
             self.app.call_from_thread(
                 bar.update_label,
-                f"{utils.get_icon('general', 'delete')[0]} {folders_to_delete[-1]}"
+                f"{utils.get_icon('general', 'delete')[0]} {folders_to_delete[-1]}",
             )
         # finished successfully
         self.app.call_from_thread(
@@ -611,7 +677,7 @@ class ProcessContainer(VerticalScroll):
         self.app.call_from_thread(bar.add_class, "done")
 
     @work(thread=True)
-    def paste_items(self, copied: list[str], cutted: list[str], dest: str = ""):
+    def paste_items(self, copied: list[str], cutted: list[str], dest: str = "") -> None:
         """
         Paste copied or cut files to the current directory
         Args:
@@ -625,7 +691,9 @@ class ProcessContainer(VerticalScroll):
         # the end, then compress, then get option by id, remove item
         if dest == "":
             dest = getcwd()
-        bar: ProgressBarContainer = self.app.call_from_thread(self.new_process_bar, classes="active")
+        bar: ProgressBarContainer = self.app.call_from_thread(
+            self.new_process_bar, classes="active"
+        )
         self.app.call_from_thread(
             bar.update_label,
             f"{utils.get_icon('general', 'paste')[0]} Getting items to paste...",
@@ -654,7 +722,9 @@ class ProcessContainer(VerticalScroll):
                 # again checks just in case something goes wrong
                 try:
                     makedirs(
-                        utils.normalise(path.join(dest, item_dict["relative_loc"], "..")),
+                        utils.normalise(
+                            path.join(dest, item_dict["relative_loc"], "..")
+                        ),
                         exist_ok=True,
                     )
                     if path.exists(path.join(dest, item_dict["relative_loc"])):
@@ -797,7 +867,9 @@ class ProcessContainer(VerticalScroll):
                 # again checks just in case something goes wrong
                 try:
                     makedirs(
-                        utils.normalise(path.join(dest, item_dict["relative_loc"], "..")),
+                        utils.normalise(
+                            path.join(dest, item_dict["relative_loc"], "..")
+                        ),
                         exist_ok=True,
                     )
                     if path.exists(path.join(dest, item_dict["relative_loc"])):
@@ -934,7 +1006,7 @@ class ProcessContainer(VerticalScroll):
                     shutil.rmtree(folder)
             except PermissionError:
                 # TODO: allow continuation and not return on error
-                self.app.notify(
+                self.notify(
                     f"Certain files in {folder} could not be moved.", severity="error"
                 )
                 self.app.call_from_thread(
@@ -958,7 +1030,7 @@ class ProcessContainer(VerticalScroll):
         )
         self.app.call_from_thread(bar.add_class, "done")
 
-    async def on_key(self, event: events.Key):
+    async def on_key(self, event: events.Key) -> None:
         if event.key in config["keybinds"]["delete"]:
             await self.remove_children(".done")
             await self.remove_children(".error")
