@@ -2,17 +2,17 @@ import os
 import platform
 import stat
 import subprocess
+from functools import lru_cache
 from os import path
-from threading import Thread
-from time import sleep
 
+import jsonschema
 import psutil
 import toml
 import ujson
+from humanize import naturalsize
 from lzstring import LZString
+from rich.console import Console
 from textual.widget import Widget
-from watchdog.events import FileSystemEventHandler
-from watchdog.observers import Observer
 
 from .maps import (
     ASCII_ICONS,
@@ -27,6 +27,7 @@ from .maps import (
 )
 
 lzstring = LZString()
+pprint = Console().print
 
 
 # What is textual reactive?
@@ -42,35 +43,21 @@ class SessionManager:
             last highlighted item. If a directory is not in the dictionary, the default is 0.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.sessionDirectories = []
         self.sessionHistoryIndex = 0
         self.sessionLastHighlighted = {}
 
 
-class ConfigError(Exception):
-    """Exception raised for config related errors.
-
-    Attributes:
-        message -- explanation of the error
-    """
-
-    def __init__(self, message):
-        self.message = message
-        super().__init__(self.message)
-
-
-state = SessionManager()
-
 config = {}
 pins = {}
 
 
 def normalise(location: str) -> str:
-    """
-    "Normalise" the path
+    """'Normalise' the path
     Args:
         location (str): The location to the item
+
     Returns:
         str: A normalised path
     """
@@ -152,10 +139,10 @@ def get_cwd_object(cwd: str, sort_order: str, sort_by: str) -> (list[dict], list
 
 
 def file_is_type(file_path: str) -> str:
-    """
-    Get a given path's type
+    """Get a given path's type
     Args:
         file_path(str): The file path to check
+
     Returns:
         str: The string that says what type it is (unknown, symlink, directory, junction or file)
     """
@@ -179,10 +166,10 @@ def file_is_type(file_path: str) -> str:
 
 
 def get_recursive_files(object_path: str) -> list[str]:
-    """
-    Get the files available at a directory recursively, regardless of whether it is a directory or not
+    """Get the files available at a directory recursively, regardless of whether it is a directory or not
     Args:
         object_path (str): The object's path
+
     Returns:
         list: A list of dictionaries, with a "path" key and "relative_loc" key
     """
@@ -214,6 +201,7 @@ def get_recursive_files(object_path: str) -> list[str]:
         return files
 
 
+@lru_cache(maxsize=128)
 def get_icon_for_file(location: str) -> list:
     """
     Get the icon and color for a file based on its name or extension.
@@ -245,6 +233,7 @@ def get_icon_for_file(location: str) -> list:
     return ICONS["file"]["default"]
 
 
+@lru_cache(maxsize=128)
 def get_icon_for_folder(location: str) -> list:
     """Get the icon and color for a folder based on its name.
 
@@ -265,12 +254,13 @@ def get_icon_for_folder(location: str) -> list:
         return ICONS["folder"]["default"]
 
 
+@lru_cache(maxsize=128)
 def get_icon(outer_key: str, inner_key: str) -> list:
-    """
-    Get an icon from double keys.
+    """Get an icon from double keys.
     Args:
         outer_key (str): The category name (general/folder/file)
         inner_key (str): The icon's name
+
     Returns:
         list[str,str]: The icon and color for the icon
     """
@@ -280,6 +270,7 @@ def get_icon(outer_key: str, inner_key: str) -> list:
         return ICONS[outer_key][inner_key]
 
 
+@lru_cache(maxsize=128)
 def get_toggle_button_icon(key: str) -> str:
     if not config["interface"]["nerd_font"]:
         return ASCII_TOGGLE_BUTTON_ICONS[key]
@@ -287,29 +278,12 @@ def get_toggle_button_icon(key: str) -> str:
         return TOGGLE_BUTTON_ICONS[key]
 
 
-def update_session_utils(directories, index, lastHighlighted={}) -> None:
-    """
-    Update the session utils with the given directories and index.
-
-    Args:
-        directories (list): List of directories in the session.
-        index (int): Current index in the session history.
-        lastHighlighted (str): The last highlighted file or directory.
-    """
-    global sessionDirectories
-    global sessionHistoryIndex
-    global sessionLastHighlighted
-    sessionDirectories = directories
-    sessionHistoryIndex = index
-    sessionLastHighlighted = lastHighlighted
-
-
-def deep_merge(d, u) -> dict:
-    """
-    Mini lodash merge
+def deep_merge(d: dict, u: dict) -> dict:
+    """Mini lodash merge
     Args:
         d (dict): old dictionary
         u (dict): new dictionary, to merge on top of d
+
     Returns:
         dict: Merged dictionary
     """
@@ -333,24 +307,75 @@ def load_config() -> None:
     if not path.exists(path.join(VAR_TO_DIR["CONFIG"], "config.toml")):
         with open(path.join(VAR_TO_DIR["CONFIG"], "config.toml"), "w") as file:
             file.write(
-                "#:schema  https://raw.githubusercontent.com/NSPC911/rovr/refs/heads/master/src/rovr/config/schema.json"
+                '#:schema  https://raw.githubusercontent.com/NSPC911/rovr/refs/heads/master/src/rovr/config/schema.json\n[theme]\ndefault = "nord"'
             )
 
     with open(path.join(path.dirname(__file__), "config/config.toml"), "r") as f:
-        template_config = toml.loads(f.read())
+        try:
+            template_config = toml.loads(f.read())
+        except toml.decoder.TomlDecodeError as e:
+            pprint(f"[red]TOML Syntax Error:\n    {e}")
+            exit(1)
 
     user_config_path = path.join(VAR_TO_DIR["CONFIG"], "config.toml")
     user_config = {}
     if path.exists(user_config_path):
-        try:
-            with open(user_config_path, "r") as f:
-                user_config_content = f.read()
-                if user_config_content:
-                    user_config = toml.loads(user_config_content)
-        except (IOError, toml.TomlDecodeError):
-            pass
+        with open(user_config_path, "r") as f:
+            user_config_content = f.read()
+            if user_config_content:
+                user_config = toml.loads(user_config_content)
     # Don't really have to consider the else part, because it's created further down
     config = deep_merge(template_config, user_config)
+    # check with schema
+    with open(path.join(path.dirname(__file__), "config/schema.json"), "r") as f:
+        schema = ujson.load(f)
+
+    # fix schema with 'required' keys
+    def add_required_recursively(node: dict) -> None:
+        if isinstance(node, dict):
+            if (
+                node.get("type") == "object" and "properties" in node
+            ) and "required" not in node:
+                node["required"] = list(node["properties"].keys())
+            for key in node:
+                add_required_recursively(node[key])
+        elif isinstance(node, list):
+            for item in node:
+                add_required_recursively(item)
+
+    add_required_recursively(schema)
+
+    try:
+        jsonschema.validate(config, schema)
+    except jsonschema.exceptions.ValidationError as exception:
+        # pprint(exception.__dict__)
+        path_str = "root"
+        if exception.path:
+            path_str = ".".join(str(p) for p in exception.path)
+        pprint(
+            f"[bold red]Config Error[/bold red] in `[bold blue]{path_str}[/bold blue]`:"
+        )
+        match exception.validator:
+            case "required":
+                pprint(f"Missing required property: {exception.message}.")
+            case "type":
+                type_error_message = (
+                    f"Invalid type: expected [yellow]{exception.validator_value}[/yellow], "
+                    f"but got [yellow]{type(exception.instance).__name__}[/yellow]."
+                )
+                pprint(type_error_message)
+            case "enum":
+                enum_error_message = (
+                    f"Invalid value [yellow]'{exception.instance}'[/yellow]. "
+                    f"Allowed values are: {exception.validator_value}"
+                )
+                pprint(enum_error_message)
+            case _:
+                pprint(f"[yellow]{exception.message}[/yellow]")
+        exit(1)
+
+
+load_config()
 
 
 def load_pins() -> dict:
@@ -598,43 +623,21 @@ if not path.exists(path.join(VAR_TO_DIR["CONFIG"], "style.tcss")):
         pass
 
 
-# watchers
-class FileEventHandler(FileSystemEventHandler):
-    @staticmethod
-    def on_modified(event):
-        if event.is_directory:
-            return
-        src_path_basename = path.basename(event.src_path)
-        if src_path_basename == "config.toml":
-            load_config()
-        elif src_path_basename == "pins.json":
-            load_pins()
-
-
-def watch_config_file() -> None:
-    event_handler = FileEventHandler()
-    observer = Observer()
-    observer.schedule(
-        event_handler, path=path.join(path.dirname(__file__), "config"), recursive=False
-    )
-    user_config_dir = VAR_TO_DIR["CONFIG"]
-    if path.exists(user_config_dir):
-        observer.schedule(event_handler, path=user_config_dir, recursive=False)
-    observer.start()
-    try:
-        while True:
-            sleep(1)
-    except Exception:
-        observer.stop()
-    observer.join()
-
-
-def start_watcher():
-    Thread(
-        target=watch_config_file,
-        daemon=True,
-    ).start()
-
-
-if __name__ == "__main__":
-    start_watcher()
+def natural_size(integer: int) -> str:
+    match config["metadata"]["filesize_suffix"]:
+        case "decimal":
+            return naturalsize(
+                value=integer, format=f"%.{config['metadata']['filesize_decimals']}f"
+            )
+        case "gnu":
+            return naturalsize(
+                value=integer,
+                gnu=True,
+                format=f"%.{config['metadata']['filesize_decimals']}f",
+            )
+        case "binary":
+            return naturalsize(
+                value=integer,
+                binary=True,
+                format=f"%.{config['metadata']['filesize_decimals']}f",
+            )
